clear all;
close all;
clc;

%task 1
%insert code
u_const = 0;
x0 = [pi/3, 0]';
T = 2;
%forward simulation
[t_1,x_1] = ode45(@(t,x) ode(t,x,u_const), [0,T],x0);
[t_1_lin,x_1_lin] = ode45(@(t,x) ode_lin(t,x,u_const), [0,T],x0);

figure;
u_1 = zeros(length(t_1),1);
plot_trajectories(t_1,x_1',u_1,'');
plot_trajectories(t_1_lin,x_1_lin',u_1,'Forward simulation');
subplot(311); legend('nonlinear','linearized');
subplot(312); legend('nonlinear','linearized');

%task 2
%insert code
c = 0.2;
A = [0 1; 1 -c];
B = [0;1];
Ts = .1;
N = 20;

sysc = ss(A,B,[],[]);
sysd = c2d(sysc,Ts);
Ad = sysd.A;
Bd = sysd.B;
Q = diag([1,.1]); R = .01;
[K,P] = dlqr(Ad,Bd,Q,R);

%simulate
x_2(:,1) = x0;
t_2(1) = 0;

for i=1:N
    %state trajectory
    x_2(:,i+1) = (Ad-Bd*K)*x_2(:,i);
    %control trajectory
    u_2(i) = -K*x_2(:,i);
    %time index
    t_2(i+1) = i*Ts;
end

figure;
plot_trajectories(t_2,x_2,u_2,'Linearized LQR');

%task 3
nx = 2;
nu = 1;
Aeq = zeros(N*(nx)+nx, N*(nx+nu)+nx);
Aeq(1:nx,1:nx) = eye(nx);
for i = 1:N
    Aeq(i*(nx)+1 : i*(nx)+nx, (i-1)*(nx+nu)+1:i*(nx+nu)+nx) = [Ad,Bd,-eye(nx)];
end
Z0 = zeros(N*(nx+nu)+nx,1);
beq = zeros(N*(nx)+nx,1);
beq(1:nx) = x0;

%solve problem
options = optimoptions('fmincon','Display','iter','Algorithm','active-set');
z_3_opt = fmincon(@(Z) cost_function(Z, Q, R, P, N),Z0,[],[],Aeq,beq,[],[],[],options);

%extract controls
u_3 = z_3_opt(nx+1:nx+nu:end);

%simulate
x_3 = x0;
t_3 = 0;
for i=1:N
    x_3(:,i+1) = Ad*x_3(:,i)+ Bd*u_3(i);
    t_3(i+1) = i*Ts;
end

figure;
plot_trajectories(t_3,x_3,u_3,'Unconstrained linear MPC');

%task 4
lb = [];
ub = [];
xmax = [inf;.8];
xmin = -xmax;
umax = inf;
umin = -umax;

for i=1:N
    ub = [ub; xmax; umax];
    lb = [lb; xmin; umin];
end
ub = [ub;xmax];
lb = [lb;xmin];

%solve problem
options = optimoptions('fmincon','Display','iter','Algorithm','active-set');
z_4_opt = fmincon(@(Z) cost_function(Z, Q, R, P, N),Z0,[],[],Aeq,beq,lb,ub,[],options);

%extract controls
u_4 = z_4_opt(3:nx+nu:end);

%simulate
x_4 = x0;
t_4 = 0;
for i=1:N
    x_4(:,i+1) = Ad*x_4(:,i)+ Bd*u_4(i);
    t_4(i+1) = i*Ts;
end

figure;
plot_trajectories(t_4,x_4,u_4,'Constrained linear MPC');

%task 5
t_5(1) = 0;
u_5 = u_4;
x_5(:,1) = x0;
for i=1:N
    uk = u_5(i);
    xk = x_5(:,i);
    [t,x] = ode45(@(t,x) ode(t,x,uk), [0 Ts], xk);
    x_5(:,i+1) = x(end,:);
    t_5(i+1) = i*Ts;
end

figure;
plot_trajectories(t_5,x_5,u_5,'Nonlinear simulation of Linear MPC');
 
%task 6
nonlcon = @(Z) nonl_constr(Z, N, Ts, nx, nu, x0);
z_6_opt = fmincon(@(Z) cost_function(Z, Q, R, P, N),Z0,[],[],[],[],lb,ub,nonlcon,options);

%extract controls
u_6 = z_6_opt(3:nx+nu:end);

%simulate
x_6(:,1) = x0;
t_6(1) = 0;
for i=1:N
    uk = u_6(i);
    xk = x_6(:,i);
    [t,x] = ode45(@(t,x) ode(t,x,uk), [0 Ts], xk);
    x_6(:,i+1) = x(end,:);
    t_6(i+1) = i*Ts;
end

figure;
plot_trajectories(t_6,x_6,u_6,'Nonlinear MPC');


